(* vpidentd - very paranoid ident daemon
   
   (C) 2002-2003 Nicholas Jay Kain <niklata@aerifal.cx>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation. 

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *)

(* START user customization *)

(* By default, it's possible to infer information about the existence of
user accounts from the responses provided by identd.  Setting this to true
will make this 'attack' impossible.  Whether this 'attack' is practically
feasible or not is something I haven't considered.

This setting violates the rfc, but should be compatible with any sane ident
client. 

Leave on if you're absurdly paranoid, turn off if you're a standards lawyer. *)
let paranoid_responses = false

(* Set value to the location of a user's .noident relative to their homedir.
File must be readable by the user that runs vpidentd. *)
let ident_filename = "/.ident"

(* END user customization *)

type ident_query = 
    Bad
  | Ports of int * int

type ident_errors =
    Unknown_error
  | Invalid_port
  | Hidden_user
  | No_user

exception Malformed
exception Close_connection

let proxy_fd = ref (Unix.stdout: Unix.file_descr)
		 
let cmp_to_mask hosti testi maski =
  let host = ref (Unix.string_of_inet_addr hosti)
  and test = ref (Unix.string_of_inet_addr testi)
  and mask = ref (Unix.string_of_inet_addr maski)
  and m = Array.create 4 0 and t = Array.create 4 0 and h = Array.create 4 0 in
    
    for i = 0 to 2 do
      m.(i) <- int_of_string (Nstring.lgrab !mask '.') ; mask := Nstring.rgrab !mask '.' ;
      t.(i) <- int_of_string (Nstring.lgrab !test '.') ; test := Nstring.rgrab !test '.' ;
      h.(i) <- int_of_string (Nstring.lgrab !host '.') ; host := Nstring.rgrab !host '.'
    done ;
    m.(3) <- int_of_string (!mask) ; t.(3) <- int_of_string (!test) ; h.(3) <- int_of_string (!host) ;

    for i = 0 to 3 do
      t.(i) <- m.(i) land t.(i) ;
      h.(i) <- m.(i) land h.(i)
    done ;

    if t.(0) = h.(0) && t.(1) = h.(1) && t.(2) = h.(2) && t.(3) = h.(3) then
      true else false

let parse_query l = 
  try
    if not (String.contains_from l 0 ',') then raise Malformed else
      let leftside = Nstring.trim (Nstring.lgrab l ',')
      and rightside = Nstring.trim (Nstring.rgrab l ',') in

      let sps = Nstring.extract_first_numeric leftside
      and dps = Nstring.extract_first_numeric rightside in

      let sp = int_of_string sps
      and dp = int_of_string dps in

	Ports (sp, dp)

  with
      Malformed -> Bad
    | Failure "int_of_string" -> Bad

(* please note that compressing the hash will greatly weaken the strength of
   md5 -- the method I use will result in something that is somewhat weaker
   than 48-bit md5, with the 'slightly' owing to numbers 0-6 being 25% more
   likely to occur than other alphanumerics and the alphanumeric mapping
   encoding slightly less than 6 bits of each byte.  however, as ident's
   responses should never be trusted, the security provided should be more
   than sufficient *)

let make_hash uname ip src dst flags = 
  let rec make_hash_string carry = function
      [] -> carry
    | Parsecfg.UID::t -> make_hash_string (uname ^ carry) t
    | Parsecfg.IP_ADDR::t -> make_hash_string (ip ^ carry) t
    | Parsecfg.S_PORT::t -> make_hash_string ((string_of_int src) ^ carry) t
    | Parsecfg.D_PORT::t -> make_hash_string ((string_of_int dst) ^ carry) t
  and compress_hash str =
    begin
      let outstr = ref "        " in 
	for i=0 to 7 do
	  (!outstr).[i] <- char_of_int ((int_of_char str.[i*2]) lxor (int_of_char str.[i*2+1]))
	done ;
	for i=0 to 7 do  (* convert to alphanumerics, one cmp can be removed *)
	  begin
	    (!outstr).[i] <- char_of_int ((int_of_char (!outstr).[i]) mod 62) ;
	    (!outstr).[i] <- char_of_int ((int_of_char (!outstr).[i]) + 48) ;
  	    if int_of_char (!outstr).[i] > 57 then
	      begin
		(!outstr).[i] <- char_of_int ((int_of_char (!outstr).[i]) + 7) ;
 		if int_of_char (!outstr).[i] > 90 then
	 	  (!outstr).[i] <- char_of_int ((int_of_char (!outstr).[i]) + 6) 
	      end
	  end 
	done ;
	!outstr
    end
  in
    compress_hash (Digest.string (make_hash_string "" flags))

exception Deny_response

let filter_unames name str =
  try
    if str = name then str else
      let info = Unix.getpwnam str in
	raise Deny_response
  with
      Not_found -> str (* no uid exists with str as name, so it's fine *)

let rec check_cfg uid sp dp name host = function
    [] -> raise Deny_response (* default is to deny *)
  | {Parsecfg.ip=i; Parsecfg.mask=m; Parsecfg.sports=(sl,sh); Parsecfg.dports=(dl,dh); Parsecfg.response=resp;} :: rem ->
      if (cmp_to_mask (Unix.inet_addr_of_string host) i m) ||
	((sp >= sl || sl = 0) && (sp <= sh || sh = 0)) ||
	((dp >= dl || dl = 0) && (dp <= dh || dh = 0))
      then 
	begin match resp with
	    Parsecfg.Deny -> raise Deny_response
	  | Parsecfg.Accept -> name
	  | Parsecfg.Spoof str -> filter_unames name str
	  | Parsecfg.Hash t -> make_hash name host sp dp t
	end
      else check_cfg uid sp dp name host rem

(* can emit exn Not_found, but should not - catch and handle it above
   can emit exn Unix_error if .ident DNE - handle above *)
let lookup_user uid sp dp host = 
  try
    let info = Unix.getpwuid uid in
    let identfile = (info.Unix.pw_dir ^ ident_filename) in
      Unix.access identfile [Unix.R_OK] ;
      let respdata = (Parsecfg.read_file identfile) in
	check_cfg uid sp dp info.Unix.pw_name host respdata
  with
      Unix.Unix_error (_, "access", _) -> raise Deny_response

let generate_response spi dpi reply =
  let sp = string_of_int spi and dp = string_of_int dpi in
    sp ^ "," ^ dp ^ ":" ^ reply ^ "\r\n"

let output_response sp dp reply fh = 
  let out = Unix.out_channel_of_descr fh in
    output_string out (generate_response sp dp reply) ;
    close_out out

let generate_error host sp dp code fh =
  let r = if paranoid_responses then "ERROR:UNKNOWN-ERROR" else
    match code with
	Unknown_error -> "ERROR:UNKNOWN-ERROR"
      | Invalid_port -> "ERROR:INVALID-PORT"
      | Hidden_user -> "ERROR:HIDDEN-USER"
      | No_user -> "ERROR:NO-USER"
  in
    Log.reply host sp dp r ;
    output_response sp dp r fh

let send_reply host sp dp u fh =
  try
    let reply = "USERID:UNIX:" ^ lookup_user u sp dp host in
      Log.reply host sp dp reply ;
      output_response sp dp reply fh
  with Not_found -> generate_error host sp dp No_user fh
    | Deny_response -> generate_error host sp dp Hidden_user fh

let dispatch line host fh = 
      match parse_query line with
	  Bad -> raise Close_connection
	| Ports (sp,dp) -> 
	    if sp > 0 && dp > 0 && sp < 65536 && dp < 65536 then
	      match Proctcp.query host sp dp with
		  Proctcp.Error -> 
		    generate_error host sp dp No_user fh
		| Proctcp.User u -> send_reply host sp dp u fh
	    else
	      generate_error host sp dp Invalid_port fh

let lookup_host fd =
  let saddy = Unix.getpeername fd in
    match saddy with
	Unix.ADDR_UNIX _ -> 
	  Log.error "I don't handle unix sockets -- something is wrong!" ;
	  raise (Failure "I don't handle unix sockets -- something is wrong.")
      | Unix.ADDR_INET (addy,_) -> Unix.string_of_inet_addr addy

let inetd_mode () = 
  let host = lookup_host Unix.stdin
  and is = Unix.in_channel_of_descr Unix.stdin in
    if Unix.getuid () = 0 || Unix.getgid () = 0 then 
      (Log.error "I refuse to bear the burdens of root.";
       raise (Failure "I refuse to bear the burdens of root."));
    try 
      dispatch (read_line ()) host Unix.stdout
    with
	Close_connection -> Log.error ("(" ^ host ^ ") sent deformed query!")
;;

inetd_mode ()
